#!/usr/bin/env python
from pprint import pprint
import websockets
from uuid import uuid4
import json
import asyncio
import math
from .binvox import Binvox
from PIL import Image
from numpy import asarray
from .palette import palette

from sklearn.neighbors import NearestNeighbors
import numpy as np

color_wheel = [
    ['Purple wool'          , 'wool 10'                  , 'Purple'              , '764b73'],
    ['Magenta wool'         , 'wool 2'                   , 'Magenta'             , 'a05077'],
    ['Pink wool'            , 'wool 6'                   , 'Pink'                , 'ce747e'],
    ['Pink hardened Clay'   , 'stained_hardened_clay 6'  , 'dark pink'           , '9c544e'],
    ['Red hardened Clay'    , 'stained_hardened_clay 14' , 'dark reddish pink'   , '8d493e'],
    ['Red wool'             , 'wool 14'                  , 'dark reddish pink'   , '9d433b'],
    ['Red mushroom block'   , 'red_mushroom_block'       , 'dark Red'            , 'ab3335'],
    ['Redstone block'       , 'redstone_block'           , 'Red'                 , 'e54547'],
    ['Orange hardened clay' , 'stained_hardened_clay 1'  , 'dark reddish orange' , 'b63b34'],
    ['Orange wool'          , 'wool 1'                   , 'Orange'              , 'd45e42'],
    ['Yellow hardened clay' , 'stained_hardened_clay 4'  , 'Yellowish orange'    , 'ca893c'],
    ['Gold block'           , 'gold_block'               , 'yellow'              , 'e3cf52'],
    ['Yellow wool'          , 'wool 4'                   , 'Yellowish green'     , 'bea831'],
    ['Melon'                , 'melon_block'              , 'Dark green'          , '6c7525'],
    ['Lime wool'            , 'wool 5'                   , 'Green'               , '64940e'],
    ['Emerald block'        , 'emerald_block'            , 'Blue green'          , '53924b'],
    ['Green wool'           , 'wool 13'                  , 'Dark green'          , '3a4827'],
    ['Prismarine bricks'    , 'prismarine 1'             , 'Dark green blue'     , '415647'],
    ['Cyan wool'            , 'wool 9'                   , 'Cyan'                , '235f6a'],
    ['Packed ice'           , 'packed_ice'               , 'Light blue gray'     , '81929c'],
    ['Light blue wool'      , 'wool 3'                   , 'Light blue'          , '686997'],
    ['Lapis lazul block'    , 'lapis_block'              , 'Dark blue'           , '3b4664'],
    ['Blue wool'            , 'wool 1'                   , 'very dark blue'      , '464165'],
    ['Blue hardened clay'   , 'stained_hardened_clay 11' , 'Almost block blue'   , '51444d'],
]


class Client:
    def __init__(self, port=3000):
        self._port = port
        self._websocket = None
        self._send_queue = []
        self._awaiting_response = {}
        self._awaiting_limit = 90
        self._nn_model = None

    def _hex2rgb(self, h):
        return tuple(int(h[i:i+2], 16) for i in (0, 2, 4))

    async def connect(self):
        uri = f"ws://0.0.0.0:{self._port}/script"
        self._websocket = await websockets.connect(uri)
        asyncio.ensure_future(self._monitor_responses())


    async def send_command(self, cmd):
        msg = {
            "header": {
                "version": 1,
                "requestId": str(uuid4()),
                "messageType": "commandRequest",
                "messagePurpose": "commandRequest",
            },
            "body": {
                "commandLine": cmd,
                "origin": {
                    "type": "player"
                }
            }
        }
        future = asyncio.get_event_loop().create_future()
        self._send_queue.append([msg, future])
        await self._consider_send_queue()
        return await future

    async def _monitor_responses(self):
        async for response_msg in self._websocket:
            response_msg = json.loads(response_msg)
            if response_msg.get('header').get('messagePurpose') == 'commandResponse':
                # ... and it's for an awaited command
                request_id = response_msg.get('header').get('requestId')
                if request_id in self._awaiting_response:
                    status = response_msg.get('body').get('statusCode')
                    # if status < 0:
                    #     import sys
                    #     print(f"fatal error {status}", file=sys.stderr)
                    # else:
                    #     future =  self._awaiting_response.get(request_id)
                    #     future.set_result(response_msg)
                    #     future.done()
                    future =  self._awaiting_response.get(request_id)
                    future.set_result(response_msg)
                    future.done()
                self._awaiting_response.pop(request_id)
                await self._consider_send_queue()
      #   // Print errors (if any)
      #   if (msg.body.statusCode < 0)
      #     console.log(awaitedQueue[msg.header.requestId].body.commandLine, msg.body.statusMessage)
      #   // ... and delete it from the awaited queue
      #   delete awaitedQueue[msg.header.requestId]
      # }
    # }



    async def _consider_send_queue(self):
        while len(self._send_queue) > 0 and len(self._awaiting_response) < self._awaiting_limit:
            item = self._send_queue.pop(0)
            import pprint
            pprint.pprint(item)
            msg =  item[0]
            future = item[1]
            self._awaiting_response[msg['header']['requestId']] = future
            print(f"Sending: {json.dumps(msg)}")
            await self._websocket.send(json.dumps(msg))
            # asyncio.ensure_future(self._recv_one_response())
            # await asyncio.sleep(0.1)
        # await self.websocket.send(json.dumps(msg))


    async def fill(self, x1, y1, x2, y2, material):
        pass

    async def draw_cyclinder(self, radius, height, material, x_offset=0, y_offset=0, z_offset=0):
        tasks = []
        for y in range(0, height):
            tasks.append(self.draw_circle(radius, material, x_offset=x_offset, y_offset=y + y_offset, z_offset=z_offset))
        results = await asyncio.gather(*tasks)


    async def draw_cone(self, radius, material, x_offset=0, y_offset=0, z_offset=0):
        tasks = []
        for y in range(0, radius):
            print(f"Runing draw_circle({radius-y}, {material}, {y + y_offset})")
            tasks.append(self.draw_circle(radius-y, material, x_offset=x_offset, y_offset=y + y_offset, z_offset=z_offset))
        results = await asyncio.gather(*tasks)

    async def draw_cube(self, x, y, size, material, hollow=False):
        pass

    async def draw_binvox(self, path, material, y_offset=0, x_offset=0, z_offset=0):
        model = Binvox.read(path, mode='dense').numpy()
        (depth, width, height) = model.shape

        tasks = []
        for x in range(0, depth):
            for z in range(0, width):
                for y in range(0, height):
                    if model[x][z][y] and (
                            x == 0
                            or x == depth - 1
                            or y == 0
                            or y == height - 1
                            or z == 0
                            or z == width - 1
                            or not model[x-1][z][y]
                            or not model[x+1][z][y]
                            or not model[x][z-1][y]
                            or not model[x][z+1][y]
                            or not model[x][z][y-1]
                            or not model[x][z][y+1]
                    ):
                        print(f"{x}, {y}, {z}")
                        tasks.append(self.send_command(f"setblock ~{x + x_offset} ~{y + y_offset} ~{z + z_offset} {material}"))

        results = await asyncio.gather(*tasks)

    async def draw_circle(self, radius, material, y_offset=0, x_offset=0, z_offset=0):
        tasks = []


        # # Figure out the key blocks we want to place
        # key_blocks = []
        # for x in range(0, radius+1):
        #     z = round(radius*math.sin(math.acos(x/radius)))
        #     key_blocks.append([x, y, z])
        #     key_blocks.append([x, y, -z])
        #     key_blocks.append([-x, y, z])
        #     key_blocks.append([-x, y, -z])

        # blocks = interpolate_blocks(key_blocks)

        last_z = None
        for x in range(0, radius+1):
            new_z = round(radius*math.sin(math.acos(x/radius)))
            zs = [new_z]
            if last_z:
                zs = range(new_z, last_z+1)

            for z in zs:
                tasks.append(self.send_command(f"setblock ~{x + x_offset} ~{y_offset} ~{z + z_offset} {material}"))
                tasks.append(self.send_command(f"setblock ~{x + x_offset} ~{y_offset} ~{-z + z_offset} {material}"))
                tasks.append(self.send_command(f"setblock ~{-x + x_offset} ~{y_offset} ~{z + z_offset} {material}"))
                tasks.append(self.send_command(f"setblock ~{-x + x_offset} ~{y_offset} ~{-z + z_offset} {material}"))
                
            last_z = new_z

        results = await asyncio.gather(*tasks)

    async def draw_monochrome_image(self, file, material="obsidian", background=None):
        image = Image.open(file)
        # convert image to numpy array
        data = asarray(image)
        height = data.shape[0]
        width = data.shape[1]
        channels = data.shape[2]

        tasks = []
        for x in range(0, width):
            for y in range(0, height):
                pixel = data[y][x]
                is_black = False
                for i in range(0, channels):
                    if pixel[i] != 255:
                        is_black=True
                        break

                if is_black:
                    tasks.append(self.send_command(f"setblock ~{x} ~{height-y} ~ {material}"))
                elif background is not None:
                    tasks.append(self.send_command(f"setblock ~{x} ~{height-y} ~ {background}"))

        await asyncio.gather(*tasks)

    def _build_nn_model(self):
        global palette
        X = np.array([x[1] for x in palette])
        # global color_wheel
        # X = np.array([self._hex2rgb(x[3]) for x in color_wheel])
        # self._nn_model = NearestNeighbors(n_neighbors=1, algorithm='ball_tree').fit(X)
        self._nn_model = NearestNeighbors(n_neighbors=1, algorithm='kd_tree').fit(X)

    async def draw_image_nn(self, file):
        global color_wheel
        image = Image.open(file)
        # convert image to numpy array
        data = asarray(image)
        height = data.shape[0]
        width = data.shape[1]
        channels = data.shape[2]

        if self._nn_model is None:
            self._build_nn_model()
        nbrs = self._nn_model

        tasks = []
        for y in range(height):
            distances, indices = nbrs.kneighbors(data[y])
            # materials = [color_wheel[i[0]][1] for i in indices]
            materials = [palette[i[0]][0] for i in indices]
            # print(f"material size is {len(materials)}")
            for x in range(width):
                tasks.append(self.send_command(f"setblock ~{x} ~{height-y} ~ {materials[x]}"))

        await asyncio.gather(*tasks)

    async def draw_palette(self):
        global palette
        tasks = []
        for i in range(len(palette)):
            tasks.append(self.send_command(f"setblock ~+{i} ~ ~ {palette[i][0]}"))
        await asyncio.gather(*tasks)



    async def draw_image(self, file):
        image = Image.open(file)
        # convert image to numpy array
        data = asarray(image)
        height = data.shape[0]
        width = data.shape[1]
        channels = data.shape[2]

        tasks = []
        for x in range(0, width):
            for y in range(0, height):
                pixel = data[y][x]

                material = self._closest_block_for_color(pixel)
                tasks.append(self.send_command(f"setblock ~{x} ~{height-y} ~ {material}"))

        await asyncio.gather(*tasks)

    async def build_pyramid(self, size, material='glowstone'):
        tasks = []
        for y in range(0, size+1):
            side = size - y
            for x in range(-side, side+1):
                tasks.append(self.send_command(f"setblock ~{x} ~{y} ~{-side} {material}"))
                tasks.append(self.send_command(f"setblock ~{x} ~{y} ~{+side} {material}"))
                tasks.append(self.send_command(f"setblock ~{-side} ~{y} ~{x} {material}"))
                tasks.append(self.send_command(f"setblock ~{+side} ~{y} ~{x} {material}"))
        await asyncio.gather(*tasks)

    def _closest_block_for_color(self, rgb):
        global color_wheel
        closest_block = None
        closest_distance = None
        # print(f"Finding closest block for point {rgb}")
        for color_record in color_wheel:
            rgb2 = self._hex2rgb(color_record[3])
            distance = math.sqrt((rgb[0]-rgb2[0])**2 + (rgb[1]-rgb2[1])**2 + (rgb[2]-rgb2[2])**2)
            if closest_block is None or distance < closest_distance:
                closest_distance = distance
                closest_block = color_record[1]

        return closest_block

    async def draw_sphere(self, radius, material):
        tasks = []

        for y in range(radius):
            segment_radius = round(math.sqrt(radius**2 - y**2))
            tasks.append(self.draw_circle(segment_radius, material, y))
            tasks.append(self.draw_circle(segment_radius, material, -y))

        results = await asyncio.gather(*tasks)



#!/usr/bin/env python

import re
import asyncio
import websockets
import json
from uuid import uuid4
import socket
import pprint

minecraftWebsocket = None
scriptWebsockets = set()


def registerScript(websocket):
    global scriptWebsockets
    scriptWebsockets.add(websocket)

def registerMinecraft(websocket):
    global minecraftWebsocket
    minecraftWebsocket = websocket


# On Minecraft, when you type "/connect localhost:3000" it creates a connection
async def mineproxy(websocket, path):
    global minecraftWebsocket, scriptWebsocket
    print("Connection to " + path + "\n")

    if "script" in path:
        registerScript(websocket)
        try:
            async for message in websocket:
                print(f"Received message from script: {message}")
                if minecraftWebsocket is not None:
                    print("Relaying the message\n")
                    await minecraftWebsocket.send(message)
        except websockets.exceptions.ConnectionClosedError:
            print("Script websocket connection closed\n")
            scriptWebsockets.remove(websocket)

    else:
        registerMinecraft(websocket)
        async for message in websocket:
            print(f"Received and discarding message from minecraft: {message}")

            closed_web_sockets = []
            for scriptWebsocket in scriptWebsockets:
                try:
                    await scriptWebsocket.send(message)
                except websockets.exceptions.ConnectionClosedError:
                    print("Received script disconnect\n")
                    closed_web_sockets.append(scriptWebsocket)

            for scriptWebsocket in closed_web_sockets:
                scriptWebsockets.remove(scriptWebsocket)

    # if minecraftWebsocket is None:
    #     minecraftWebsocket = websocket
    #     print('Minecraft connected')
    # else:
    #     scriptWebsocket = websocket
    #     print('Script connected')

    # # Tell Minecraft to send all chat messages. Required once after Minecraft starts
    # await websocket.send(
    #     json.dumps({
    #         "header": {
    #             "version": 1,                     # We're using the version 1 message protocol
    #             "requestId": str(uuid4()),        # A unique ID for the request
    #             "messageType": "commandRequest",  # This is a request ...
    #             "messagePurpose": "subscribe"     # ... to subscribe to ...
    #         },
    #         "body": {
    #             "eventName": "PlayerMessage"
    #         },
    #     }))

    # print("Awaiting message\n")
    # async for message in websocket:
    #     print("Received message: ")
    #     pprint.pprint(message)

    # print("Done waiting for messages\n")


    # def send(cmd):
    #     # TODO
    #     pass

    # # def draw_pyramid(size):
    # #     # TODO
    # #     pass

    # # try:
    # #     # When MineCraft sends a message (e.g. on player chat), print it.
    # #     async for msg in websocket:
    # #         msg = json.loads(msg)
    # #         if msg['body']['eventName'] == 'PlayerMessage':
    # #             match = re.match(r'^pyramid (\d+)', msg['body']['properties']['Message'],
    # #                              re.IGNORECASE)
    # #             if match:
    # #                 draw_pyramid(int(match.group(1)))
    # # except websockets.exceptions.ConnectionClosedError:
    # #     print('Disconnected from MineCraft')


start_server = websockets.serve(mineproxy, host="0.0.0.0", port=3000)

host_name = socket.gethostname()
host_ip = socket.gethostbyname(host_name)
print(f'Ready. On MineCraft chat, type /connect {host_ip}:3000')

asyncio.get_event_loop().run_until_complete(start_server)
asyncio.get_event_loop().run_forever()



# This is temporarily copied from binvox since that module is broken.  Here we change np.bool to bool
# Source https://github.com/faridyagubbayli/binvox
from typing import Tuple
import numpy as np

class Binvox(object):
    """ Holds a binvox model.
    data is either a three-dimensional numpy boolean array (dense representation)
    or a two-dimensional numpy float array (coordinate representation).

    dims, translate and scale are the model metadata.

    dims are the voxel dimensions, e.g. [32, 32, 32] for a 32x32x32 model.

    scale and translate relate the voxels to the original model coordinates.
    """
    def __init__(self, data=None, dims=None, translate=None, scale=None, axis_order=None, mode=None):
        self.data           = data
        self.dims           = dims
        self.translate      = translate
        self.scale          = scale
        self.axis_order     = axis_order
        self.mode           = mode
        assert axis_order in ('xzy', 'xyz')

    @staticmethod
    def read(filepath, mode: str, fix_coords=True):
        assert mode in ['dense', 'sparse'], 'Mode should be either `dense` or `sparse`.'

        with open(filepath, 'rb') as fp:
            dims, translate, scale = Binvox.read_header(fp)
            raw_data = np.frombuffer(fp.read(), dtype=np.uint8)

        methods = {
            'dense': Binvox._parse_dense,
            'sparse': Binvox._parse_sparse,
        }
        method = methods[mode]

        data, axis_order = method(raw_data, dims, fix_coords)
        return Binvox(data, dims, translate, scale, axis_order, mode)

    @staticmethod
    def read_header(fp):
        """
            Read the binvox file header
        :param fp: File pointer
        :return: Tuple(Dims, Translate, Scale)
        """
        """ Read binvox header. Mostly meant for internal use.
        """
        line = fp.readline().strip()
        if not line.startswith(b'#binvox'):
            raise IOError('Not a binvox file')
        dims = list(map(int, fp.readline().strip().split(b' ')[1:]))
        translate = list(map(float, fp.readline().strip().split(b' ')[1:]))
        scale = list(map(float, fp.readline().strip().split(b' ')[1:]))[0]
        line = fp.readline()
        return dims, translate, scale

    @staticmethod
    def _parse_dense(raw_data, dims, fix_coords):
        """ Read binary binvox format as array.

        Returns the model with accompanying metadata.

        Voxels are stored in a three-dimensional numpy array, which is simple and
        direct, but may use a lot of memory for large models. (Storage requirements
        are 8*(d^3) bytes, where d is the dimensions of the binvox model. Numpy
        boolean arrays use a byte per element).

        Doesn't do any checks on input except for the '#binvox' line.
        """
        # if just using reshape() on the raw data:
        # indexing the array as array[i,j,k], the indices map into the
        # coords as:
        # i -> x
        # j -> z
        # k -> y
        # if fix_coords is true, then data is rearranged so that
        # mapping is
        # i -> x
        # j -> y
        # k -> z
        values, counts = raw_data[::2], raw_data[1::2]
        data = np.repeat(values, counts).astype(bool)
        data = data.reshape(dims)
        if fix_coords:
            # xzy to xyz TODO the right thing
            data = np.transpose(data, (0, 2, 1))
            axis_order = 'xyz'
        else:
            axis_order = 'xzy'
        return data, axis_order

    @staticmethod
    def _parse_sparse(raw_data, dims, fix_coords):
        """ Read binary binvox format as coordinates.

        Returns binvox model with voxels in a "coordinate" representation, i.e.  an
        3 x N array where N is the number of nonzero voxels. Each column
        corresponds to a nonzero voxel and the 3 rows are the (x, z, y) coordinates
        of the voxel.  (The odd ordering is due to the way binvox format lays out
        data).  Note that coordinates refer to the binvox voxels, without any
        scaling or translation.

        Use this to save memory if your model is very sparse (mostly empty).

        Doesn't do any checks on input except for the '#binvox' line.
        """

        values, counts = raw_data[::2], raw_data[1::2]

        sz = np.prod(dims)
        index, end_index = 0, 0
        end_indices = np.cumsum(counts)
        indices = np.concatenate(([0], end_indices[:-1])).astype(end_indices.dtype)

        values = values.astype(bool)
        indices = indices[values]
        end_indices = end_indices[values]

        nz_voxels = []
        for index, end_index in zip(indices, end_indices):
            nz_voxels.extend(range(index, end_index))
        nz_voxels = np.array(nz_voxels)
        # TODO are these dims correct?
        # according to docs,
        # index = x * wxh + z * width + y; // wxh = width * height = d * d

        x = nz_voxels / (dims[0]*dims[1])
        zwpy = nz_voxels % (dims[0]*dims[1]) # z*w + y
        z = zwpy / dims[0]
        y = zwpy % dims[0]
        if fix_coords:
            data = np.vstack((x, y, z))
            axis_order = 'xyz'
        else:
            data = np.vstack((x, z, y))
            axis_order = 'xzy'

        return np.ascontiguousarray(data), axis_order

    def numpy(self):
        return self.data

    def write(self, filepath):
        """ Write binary binvox format.

        Note that when saving a model in sparse (coordinate) format, it is first
        converted to dense format.

        Doesn't check if the model is 'sane'.

        """
        if self.mode == 'sparse':
            self.to_sparse()
        dense_data = self.numpy()

        with open(filepath, 'wb') as fp:
            header = "#binvox 1\n" + \
                     'dim '+' '.join(map(str, self.dims))+'\n' + \
                     'translate '+' '.join(map(str, self.translate))+'\n' + \
                     'scale '+str(self.scale)+'\n' + \
                     'data\n'
            fp.write(str.encode(header))

            if self.axis_order == 'xzy':
                voxels_flat = dense_data.flatten()
            elif self.axis_order == 'xyz':
                voxels_flat = np.transpose(dense_data, (0, 2, 1)).flatten()
            else:
                raise NotImplementedError('Unsupported voxel model axis order')

            # keep a sort of state machine for writing run length encoding
            voxels_compressed = self.compress_flat_voxels(voxels_flat)
            BYTE_SIZE = 1
            for entry in voxels_compressed:
                fp.write(entry.to_bytes(BYTE_SIZE, 'big'))

    def __copy__(self):
        data        = self.data.copy()
        dims        = self.dims[:]
        translate   = self.translate[:]
        return Binvox(data, dims, translate, self.scale, self.axis_order)

    def to_sparse(self):
        if self.mode == 'sparse':
            return

        """ From dense representation to sparse (coordinate) representation.
        No coordinate reordering.
        """
        assert self.data.ndim == 3, 'Data is wrong shape; should be 3D array.'
        self.data = np.asarray(np.nonzero(self.data), np.int)
        self.mode = 'sparse'

    def to_dense(self):
        if self.mode == 'dense':
            return

        assert self.data.ndim == 2 and self.data.shape[0] == 3, \
            'Data is wrong shape; should be 3xN array.'

        if np.isscalar(self.dims):
            self.dims = [self.dims]*3

        dims = np.atleast_2d(self.dims).T
        # truncate to integers
        xyz = self.data.astype(np.int)
        # discard voxels that fall outside dims
        valid_ix = ~np.any((xyz < 0) | (xyz >= dims), 0)
        xyz = xyz[:,valid_ix]
        out = np.zeros(dims.flatten(), dtype=bool)
        out[tuple(xyz)] = True

        self.data = out
        self.mode = 'dense'

    def transform_coord(self, voxel_coord: Tuple[int, int, int]) -> Tuple[float, float, float]:
        """
            Transform voxel coordinates i, j, k to original coordinates x, y, z:
        :param voxel_coord: Voxel coordinate
        :return: Original coordinate
        """
        voxel_coord = np.array(voxel_coord, dtype=float)
        coord = voxel_coord / self.dims
        coord = self.scale * coord + self.translate
        return coord

    def compress_flat_voxels(self, voxels_flat):
        # keep a sort of state machine for writing run length encoding
        state = voxels_flat[0]
        counter = 0
        compressed = []
    
        for c in voxels_flat:
            if c == state:
                counter += 1
                # if counter hits max, dump
                if counter == 255:
                    compressed += [int(state), counter]
                    counter = 0
            else:
                # if switch state, dump
                compressed += [int(state), counter]
                state = c
                counter = 1
        # flush out remainders
        if counter > 0:
            compressed += [int(state), counter]
    
        return compressed    
